#include "conf.hh"
#ifdef HAVE_DJVU

#include <config.h>
#include <GException.h>
#include <GSmartPointer.h>
#include <GRect.h>
#include <GPixmap.h>
#include <GBitmap.h>
#include <DjVuImage.h>
#include <DjVuDocument.h>
#include <DjVuPalette.h>
#include <GOS.h>
#include <ByteStream.h>
#include <DjVuMessage.h>

#include <locale.h>
#include <stdio.h>
#include <stdlib.h>

#include <GL/gl.h>


#include "pixmap.hh"
#include "document.hh"
#include "global.hh"

class DJVUDocument : public ViewableDocument {
    DJVUDocumentTree *parent;
    int page;
  public:
    DJVUDocument(DJVUDocumentTree *parent, int page);
    Pixmap *get_pixmap();
    void init();
};

static int _roundup(int x)
{
  int r=16;
  while(r<x)
    r<<=1;
  return r;
}

class DJVUPixmap : public Pixmap {
  public:
    DJVUPixmap(DJVUDocumentTree *parent, int page);
};

DJVUDocument::DJVUDocument(DJVUDocumentTree *parent, int page) : parent(parent), page(page)
{
  printf("d");
}
void DJVUDocument::init()
{
  if(!flags&DOCUMENT_TREE_INITED) {
    flags|=DOCUMENT_TREE_INITED;
  }
}
Pixmap *DJVUDocument::get_pixmap()
{
  if(p)
    return p;
  p=new DJVUPixmap(parent, page);
  printf("page loaded, p=%x, p->data=%x\n", p, p->data);
  return p;
}

DJVUDocumentTree::DJVUDocumentTree(char *uri0)
{
  uri=strdup(uri0);
  children=new vector<DocumentTree *>;
}
void DJVUDocumentTree::init()
{
  if(!flags&DOCUMENT_TREE_INITED) {
    flags|=DOCUMENT_TREE_INITED;

    printf("loading djvu images\n");

    djvu_programname("ddjvu"); 

    doc=DjVuDocument::create_wait(GURL::Filename::UTF8(uri));
    if (! doc->wait_for_complete_init())
      G_THROW( ERR_MSG("ddjvu.failed") );
    int p=doc->get_pages_num();
    printf("%d pages found\n", p);
    for(int i=0; i<p; i++)
      children->push_back(new DJVUDocument(this, i));
  }
}
DJVUPixmap::DJVUPixmap(DJVUDocumentTree *doc, int page_num)
{
    unsigned long start, stop;
    int flag_verbose=1;
    printf("loading page %d\n", page_num);
    w=1; h=1;
    flip=true;

    GP<GPixmap> &pm=*new GP<GPixmap>;
    GP<GBitmap> &bm=*new GP<GBitmap>;

    double flag_scale = 200;
    int    flag_size = -1;
    int    flag_aspect = -1;
    int    flag_subsample = -1;
    int    flag_segment = -1;
    int    flag_mode = 0;
    bool   flag_orig_size = false;
    bool   flag_rle = false;

    GRect fullrect;
    GRect segmentrect;


    // Create DjVuImage
    start=GOS::ticks();
    GP<DjVuImage> dimg=doc->doc->get_page(page_num);
    if (!dimg || ! dimg->wait_for_complete_decode() )
      G_THROW( ERR_MSG("ddjvu.failed") );
    stop=GOS::ticks();

    // Verbose
    if (flag_verbose)
    {
      DjVuFormatErrorUTF8( "%s",(const char*)dimg->get_long_description() );
      DjVuFormatErrorUTF8( "%s\t%lu", ERR_MSG("ddjvu.decode"), stop - start);
    }

    // Check
    DjVuInfo *info = dimg->get_info();
    int colorp = dimg->is_legal_photo();
    int blackp = dimg->is_legal_bilevel();
    int compoundp = dimg->is_legal_compound();
    if (flag_verbose)
    {
      if (compoundp)
	DjVuWriteError( ERR_MSG("ddjvu.compound") );
      else if (colorp)
	DjVuWriteError( ERR_MSG("ddjvu.photo") );
      else if (blackp)
	DjVuWriteError( ERR_MSG("ddjvu.bilevel") );
      // Without included files
      DjVuFormatErrorUTF8( "%s\t%4.1f", ERR_MSG("ddjvu.memory"), 
	  (double)(dimg->get_djvu_file()->get_memory_usage())/1024 );
    }    
    if (!compoundp && !colorp && !blackp)
    { 
      DjVuPrintErrorUTF8("%s","Warning: This is not a well formed DjVu image\n");
      if (!info)
	G_THROW( ERR_MSG("ddjvu.no_info") ); 
    }

    // Setup rectangles
    if (flag_size<0 && flag_scale<0 && flag_subsample<0 && ! flag_orig_size)
      flag_scale = 100;
    if (flag_subsample>0)
      flag_scale = (double) info->dpi / flag_subsample;
    if(flag_orig_size)
      fullrect=GRect(0, 0, dimg->get_width(), dimg->get_height());

    //printf("gw=%d, gh=%d, dpi=%d\n", dimg->get_width(), dimg->get_height(), info->dpi);
    if (flag_scale>0)
    {
      int w = (int)(dimg->get_width() * flag_scale / info->dpi);
      int h = (int)(dimg->get_height() * flag_scale / info->dpi);
      if (w<1) w=1;
      if (h<1) h=1;
      fullrect = GRect(0,0, w, h);
    }
    if (flag_aspect <= 0)
    {
      int w = fullrect.width();
      int h = fullrect.height();
      double dw = (double)dimg->get_width()/ w;
      double dh = (double)dimg->get_height()/h;
      if (dw > dh) 
	h = (int)(dimg->get_height()/dw);
      else
	w = (int)(dimg->get_width()/dh);
      fullrect = GRect(0,0, w, h);
    }
    if (flag_segment < 0)
      segmentrect = fullrect;

    // Render
    start = GOS::ticks();
    switch(flag_mode)
    {
      case 's':
	bm = dimg->get_bitmap(segmentrect, fullrect);
	break;
      case 'f':
	pm = dimg->get_fg_pixmap(segmentrect, fullrect);
	if (! pm)
	  bm = dimg->get_bitmap(segmentrect, fullrect);
	break;
      case 'b':
	pm = dimg->get_bg_pixmap(segmentrect, fullrect);
	break;
      default:
	pm = dimg->get_pixmap(segmentrect, fullrect);
	if (! pm)
	  bm = dimg->get_bitmap(segmentrect, fullrect);
	break;
    }
    stop = GOS::ticks();
    if (flag_verbose)
    {
      DjVuFormatErrorUTF8( "%s\t%lu", ERR_MSG("ddjvu.render"), stop - start);
    }

    // Save image
    if (pm) 
    {
      w=pm->columns();
      h=pm->rows();
      colorspace=GL_BGR;
      compute_alloc_size();
      GPixel *p=(*pm)[0];
      if(alloc_size>w*h*components) {
	size_t z;
	data=(unsigned char *)pm->take_data(z);
      } else {
	data=(unsigned char *)malloc(alloc_size);
	memcpy(data, (unsigned char *)p, w*h*components);
      }
    }
    else if (bm) 
    {
      w=bm->columns();
      h=bm->rows();
      colorspace=GL_LUMINANCE;
      compute_alloc_size();
      unsigned char *b=(*bm)[0];
      if(alloc_size>w*h*components) {
	size_t z;
	data=(unsigned char *)bm->take_data(z);
      } else {
	data=(unsigned char *)malloc(alloc_size);
	memcpy(data, b, w*h);
      }
    }
    else
      G_THROW( ERR_MSG("ddjvu.cant_render") );
  image_valid=true;
}
#endif
