#include "conf.hh"
#include "document.hh"
#include "glpixmap.hh"
#include "glrasterpixmap.hh"
#include "util.hh"

#ifdef __HAVE_DJVU
#include "config.h"
#endif

DocumentTree::DocumentTree()
{
  flags=0, children=NULL;
}
DocumentTree::~DocumentTree()
{
  if(children) {
    for(int i=0; i<children->size(); i++)
      delete (*children)[i];
    delete children;
  }
}
ViewableDocument *DocumentTree::next(vector<int> *path, int elem)
{
  printf("next: %d, size=%d, children: %d\n", elem, path->size(), children->size());
  if(path->size()==elem)
    if(children->size()==0)
      return NULL;
    else
      path->push_back(0);
  while((*path)[elem]<children->size()) {
    (*children)[(*path)[elem]]->init();
    printf("calling child %d\n", (*path)[elem]);
    ViewableDocument *r=(*children)[(*path)[elem]]->next(path, elem+1);
    if(r)
      return r;
    (*path)[elem]++;
  }
  path->pop_back();
  return NULL;
}
ViewableDocument *DocumentTree::prev(vector<int> *path, int elem)
{
  printf("prev: %d\n", elem);
  if(path->size()==elem&&children->size()>0)
    path->push_back(children->size()-1);
  while((*path)[elem]>=0) {
    (*children)[(*path)[elem]]->init();
    ViewableDocument *r=(*children)[(*path)[elem]]->prev(path, elem+1);
    if(r)
      return r;
    (*path)[elem]--;
  }
  path->pop_back();
  return NULL;
}

void ViewableDocument::init()
{
  if(!(flags&DOCUMENT_TREE_INITED))
    flags|=DOCUMENT_TREE_INITED;
}

Pixmap *ViewableDocument::get_pixmap()
{
  if(p)
    return p;
  printf("creating pixmap for '%s'\n", uri);
  int l=strlen(uri);
  const char *ext=get_extension(uri);
  PixmapFactory **pf=pixmap_factories;
  p=NULL;
  printf("extension: %s\n", ext);
  while(*pf) {
    printf("Trying plugin %s\n", (*pf)->get_name());
    if((*pf)->accept_extension(ext)) {
      p=(*pf)->create(uri);
      if(p&&p->image_valid) {
	printf("succeeded\n");
	break;
      }
    }
    ++pf;
  }

  if(p&&!p->image_valid) {
    p->unref();
    p=NULL;
    printf("delete pixmap\n");
  }
  if(!p)
    printf("No Image plugin found for file %s\n", uri);
  return p;
}

ViewableDocument::ViewableDocument()
{
  uri=NULL;
  p=NULL, glp=NULL;
}
ViewableDocument::ViewableDocument(char *uri0)
{
  uri=strdup(uri0);
  p=NULL, glp=NULL;
}
GLPixmap *ViewableDocument::get_glp()
{
  if(glp)
    return glp;
  printf("vd::get_glp getting pixmap\n");
  if((p=get_pixmap())!=NULL) {
    printf("using pixmap with data %x\n", p->data);
    glp=new GLSimplePixmap(p);
    if(!glp->is_loaded()) {
      delete glp;
      glp=NULL;
    }
    if(p) {
      p->unref();
      p=NULL;
      printf("delete pixmap\n");
    }
  }
  printf("vd::get_glp exiting\n");
  return glp;
}
ViewableDocument *ViewableDocument::next(vector<int> *path, int elem)
{
  printf("vd::next, elem=%d, psize=%d, p=%x\n", elem, path->size(), p);
  //if(path->size()==elem&&!p)
  //  return NULL;
  if(path->size()==elem) {
    printf("show\n");
    path->push_back(0);
    return this;
  } else {
    printf("exit\n");
    path->pop_back();
    return NULL;
  }
}
ViewableDocument *ViewableDocument::prev(vector<int> *path, int elem)
{
  if(path->size()==elem) {
    path->push_back(0);
    return this;
  } else {
    path->pop_back();
    return NULL;
  }
}
void ViewableDocument::del_glp()
{
  if(glp) {
    delete glp;
    glp=NULL;
  }
}
ViewableDocument::~ViewableDocument()
{
  if(glp)
    delete glp;
  if(p)
    delete p;
  if(uri)
    free(uri);
}
URIContainerTree::URIContainerTree()
{
  children=new vector<DocumentTree *>;
}
void DocumentTree::init()
{
  if(!(flags&DOCUMENT_TREE_INITED))
    flags|=DOCUMENT_TREE_INITED;
}
void URIContainerTree::add(char *uri)
{
  if(0==1);
#ifdef HAVE_DJVU
  else if(has_mimetype(uri, "djvu"))
    children->push_back(new DJVUDocumentTree(uri));
#endif
//#ifdef HAVE_TIFF
//  else if(has_mimetype(uri, "tiff"))
//    children->push_back(new TIFFDocumentTree(uri));
//#endif
#ifdef HAVE_GS
  else if(has_mimetype(uri, "ps")||has_mimetype(uri, "pdf"))
    children->push_back(new PSDocumentTree(uri));
#endif
  else
    children->push_back(new ViewableDocument(uri));
}
