#ifndef __DOCUMENT_HH
#define __DOCUMENT_HH
#include "conf.hh"

#ifdef HAVE_DJVU
#include "config.h"
#include <DjVuDocument.h>
#endif

#include <vector>
#include "pixmap.hh"
#include "glpixmap.hh"


#define DOCUMENT_TREE_INITED       1
#define PSDOCTREE_TMPURI	  32

using namespace std;

class ViewableDocument;

class DocumentTree {
  protected:
    unsigned long flags;
    vector<DocumentTree *> *children;
  public:
    virtual void init();
    virtual ViewableDocument *next(vector<int> *path, int elem);
    virtual ViewableDocument *prev(vector<int> *path, int elem);
    DocumentTree();
    virtual ~DocumentTree();
};

class URIContainerTree : public DocumentTree {
  public:
    URIContainerTree();
    void add(char *uri);
};

class ViewableDocument : public DocumentTree {
  protected:
    Pixmap *p;
  public:
    GLPixmap *glp;
    char *uri;
    ViewableDocument(char *uri);
    ViewableDocument();
    virtual Pixmap *get_pixmap();
    void del_glp();
    GLPixmap *get_glp();
    ViewableDocument *next(vector<int> *path, int elem);
    ViewableDocument *prev(vector<int> *path, int elem);
    void init();
    ~ViewableDocument();
};

#ifdef HAVE_DJVU
class DJVUDocumentTree : public DocumentTree {
  char *uri;
  public:
  GP<DjVuDocument> doc;
    DJVUDocumentTree(char *uri0);
    void init();
};

#endif

#ifdef HAVE_TIFF
typedef struct tiff TIFF;
class TIFFDocumentTree : public DocumentTree {
    char *uri;
  public:
    TIFF *doc;
    TIFFDocumentTree(char *uri0);
    void init();
};

#endif

#ifdef HAVE_GS
class PSDocumentTree : public DocumentTree {
    char *uri;
    int child;
  public:
    char *psfile;
    int rfd, wfd;
    PSDocumentTree(char *uri0);
    ~PSDocumentTree();
    void init();
};

#endif

#endif
