#include "glext.hh"
#include <GL/gl.h>
#include <GL/glx.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

GLExt *glext;

bool _is_extension_supported(char *s)
{
  const GLubyte *exts=glGetString(GL_EXTENSIONS);
  bool supp=(strstr((const char *)exts, s)!=NULL);
  fprintf(stderr, "%s%s supported\n", s, supp?"":" not");
  return supp;
}

typedef void (*fn)();

fn _get_address(const char *s)
{
  fn f=glXGetProcAddressARB((const GLubyte *) s);
  if(!f) {
    fprintf(stderr, "Couldn't get address of %s!\n", s);
    exit(1);
  }
  return f;
}

GLExt::GLExt()
{
  nv_tex_rect_supported             =_is_extension_supported("GL_NV_texture_rectangle");
  arb_multitexture_supported        =_is_extension_supported("GL_ARB_multitexture");
  nv_register_combiners2_supported  =_is_extension_supported("GL_NV_register_combiners2");
  ext_pixel_buffer_object_supported =_is_extension_supported("GL_EXT_pixel_buffer_object");
  nv_pixel_data_range_supported     =_is_extension_supported("GL_NV_pixel_data_range");

  use_yuv=nv_register_combiners2_supported;
  use_yuv_complex=(nv_register_combiners2_supported&&ext_pixel_buffer_object_supported);

  if(arb_multitexture_supported) {
    glMultiTexCoord2f=(void (*)(GLenum, GLfloat, GLfloat)) _get_address("glMultiTexCoord2fARB");
    glActiveTexture=(void (*)(GLenum)) _get_address("glActiveTextureARB");
  }

  if(nv_register_combiners2_supported) {
    glCombinerParameterfv=(void(*) (GLenum, const GLfloat *))                  _get_address("glCombinerParameterfvNV");
    glCombinerParameteri=(void(*) (GLenum, GLint))                             _get_address("glCombinerParameteriNV");
    glCombinerInput=(void(*) (GLenum, GLenum, GLenum, GLenum, GLenum, GLenum)) _get_address("glCombinerInputNV");
    glCombinerOutput=(void(*) (GLenum, GLenum, GLenum, GLenum, GLenum, GLenum,
	GLenum, GLboolean, GLboolean, GLboolean))                              _get_address("glCombinerOutputNV");
    glFinalCombinerInput=(void(*) (GLenum, GLenum, GLenum, GLenum))            _get_address("glFinalCombinerInputNV");
    glSecondaryColor3f=(void(*) (GLfloat, GLfloat, GLfloat))                   _get_address("glSecondaryColor3f");
  }

  if(ext_pixel_buffer_object_supported) {
    glGenBuffers=(void(*) (GLsizei, GLuint *)) _get_address("glGenBuffersARB");
    glBindBuffer=(void(*) (GLenum, GLuint)) _get_address("glBindBufferARB");
    glBufferData=(void(*) (GLenum, GLsizei , const GLvoid *, GLenum)) _get_address("glBufferDataARB");
    glMapBuffer =(void *(*) (GLenum, GLenum)) _get_address("glMapBufferARB");
    glUnmapBuffer=(GLboolean(*)(GLenum)) _get_address("glUnmapBufferARB");
    glDeleteBuffers=(void(*) (GLsizei, const GLuint *)) _get_address("glDeleteBuffersARB");
  }

  if(nv_pixel_data_range_supported) {
    glPixelDataRange=(void (*)(GLenum, GLsizei, GLvoid *)) _get_address("glPixelDataRangeNV");
  }
}
