#include <GL/glu.h>
#include <GL/glext.h>

#include "glpixmap.hh"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "glext.hh"

#define PIXEL_UNPACK_BUFFER_EXT                      0x88EC
#define GL_STREAM_DRAW                    0x88E0
#define GL_WRITE_ONLY                     0x88B9
#define BUFFER_OFFSET(i) ((char *)NULL + (i))
#define GL_PIXEL_UNPACK_BUFFER_BINDING_EXT 0x88EF

GLPixmap::~GLPixmap()
{
}
bool GLPixmap::is_loaded()
{
  return loaded;
}
void GLPixmap::set_ratio(double r)
{
  ratio=r;
}
double GLPixmap::get_ratio()
{
  return ratio;
}
GLSimplePixmap::GLSimplePixmap(Pixmap *p)
{
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  colorspace=p->colorspace;
  flip=p->flip;
  rotate=0;
  if(colorspace==PIXMAP_COLORSPACE_YUV_COMPLEX)
    glGenTextures(3, texture);
  else
    glGenTextures(1, texture);
  texture_target=GL_TEXTURE_2D;
  int rw, rh;
  if(glext->nv_tex_rect_supported) {
    texture_target=GL_TEXTURE_RECTANGLE_NV;
    glEnable(GL_TEXTURE_RECTANGLE_NV);
    rw=p->w, rh=p->h;
  } else {
    glPixelStorei(GL_UNPACK_ROW_LENGTH, p->w);
    glPixelStorei(GL_UNPACK_IMAGE_HEIGHT, p->h);
    rw=roundup(p->w), rh=roundup(p->h);
  }
  glBindTexture(texture_target, *texture);
  glTexParameteri(texture_target, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(texture_target, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(texture_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(texture_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  GLenum out_space;
  if(p->colorspace==GL_LUMINANCE||p->colorspace==GL_BITMAP)
    out_space=GL_LUMINANCE;
  else
    out_space=GL_RGB;
  printf("[glsimplepixmap] datap: %x, data: %d, space: %s\n", p->data, p->alloc_size, get_color_space_string(colorspace));
  GLenum format, type;
  if(p->colorspace==GL_BITMAP)
    format=GL_COLOR_INDEX;
  else if(p->colorspace==PIXMAP_COLORSPACE_YUV)
    format=GL_RGB;
  else
    format=p->colorspace;
  if(p->colorspace==GL_BITMAP)
    type=GL_BITMAP;
  else
    type=GL_UNSIGNED_BYTE;
  if(!glext->ext_pixel_buffer_object_supported) {
    texture_is_buffer_object=false;
    if(p->use_data) {
      printf("texture target is %x, w=%d, h=%d, out_space=%s, format=%s, type=%x, data=%x\n", (texture_target), rw,
	  rh, get_color_space_string(out_space), get_color_space_string(format), (type), p->data);
      glTexImage2D(texture_target, 0, out_space, rw, rh, 0, format, type, p->data);
    }
  } else {
    {
      texture_is_buffer_object=true;
      if(colorspace==PIXMAP_COLORSPACE_YUV_COMPLEX) {
	char *pboMemory[3];
	out_space=GL_LUMINANCE;
	GLuint buf;
	int alloc_size[3];
	for(int i=0; i<3; i++)
	  alloc_size[i]=p->alloch_per_channel[i]*p->allocw_per_channel[i]*p->iterations;
	printf("w=%d, alw=%d\n", p->w, p->allocw_per_channel[0]);
	unsigned char **data[3];
	printf("w: %d, h: %d, ah: %d %d %d, aw: %d %d %d, i: %d\n", p->w, p->h,
	    p->alloch_per_channel[0], p->alloch_per_channel[1],
	    p->alloch_per_channel[2], p->allocw_per_channel[0],
	    p->allocw_per_channel[1], p->allocw_per_channel[2], p->iterations);
	for(int i=0; i<3; i++) {
	  data[i]=(unsigned char **)malloc(sizeof(unsigned char*)*p->alloch_per_channel[i]);
	  for(int j=0; j<p->alloch_per_channel[i]; j++) {
	    data[i][j]=(unsigned char *)malloc(p->allocw_per_channel[i]);
	  }
	}
	glext->glGenBuffers(1, &buf);
	glext->glBindBuffer(PIXEL_UNPACK_BUFFER_EXT, buf);
	glext->glBufferData(PIXEL_UNPACK_BUFFER_EXT, alloc_size[0]+alloc_size[1]+alloc_size[2], NULL, GL_STREAM_DRAW);
	pboMemory[0]=(char *)glext->glMapBuffer(PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY);
	pboMemory[1]=&pboMemory[0][alloc_size[0]];
	pboMemory[2]=&pboMemory[1][alloc_size[1]];
	char *porig[3];
	for(int i=0; i<3; i++)
	  porig[i]=pboMemory[i];

	for(int i=0; i<p->iterations; i++) {
	  p->read_scanlines(data);
	  for(int ch=0; ch<3; ch++)
	    for(int i=0; i<p->alloch_per_channel[ch]; i++) {
	      memcpy(pboMemory[ch], data[ch][i], p->allocw_per_channel[ch]);
	      pboMemory[ch]=&(pboMemory[ch])[p->allocw_per_channel[ch]] ;
	    }
	}

	for(int i=0; i<3; i++) {
	  for(int j=0; j<p->alloch_per_channel[i]; j++) {
	    free(data[i][j]);
	  }
	  free(data[i]);
	}

	/*unsigned char *mem=(unsigned char *) pboMemory;

	  for(int i=0; i<p->iterations; i++) {
	  for(int i=0; i<p->alloch_per_channel[0]; i++)
	  data[0][i]=&mem[i*p->allocw_per_channel[0]];
	  p->read_scanlines(data);
	  mem+=p->alloch_per_channel[0]*p->allocw_per_channel[0];
	  }
	 */
	glext->glUnmapBuffer(PIXEL_UNPACK_BUFFER_EXT);
	format=GL_LUMINANCE;
	printf("out_space=%s, rw=%d, rh=%d, format=%s\n", get_color_space_string(out_space), rw, rh, get_color_space_string(format));
	for(int i=0; i<3; i++) {
	  glBindTexture(texture_target, texture[i]);
	  glTexImage2D(texture_target, 0, out_space, p->allocw_per_channel[i], p->alloch_per_channel[i]*p->iterations,
	      0, format, type, (GLvoid *)((i==0)?0:(i==1)?(alloc_size[0]):(alloc_size[0]+alloc_size[1])));
	}
	glext->glDeleteBuffers(1, &buf);
      } else {
	printf("[comps=%d]\n", p->components);
	void *pboMemory, *texData=p->data;
	//GLuint buf;
	glext->glGenBuffers(1, &buf);
	glext->glBindBuffer(PIXEL_UNPACK_BUFFER_EXT, buf);
	int alloc_size=p->alloc_size;
	if(p->components==3)
	  alloc_size=rw*rh*4;

	glext->glBufferData(PIXEL_UNPACK_BUFFER_EXT, alloc_size, NULL, GL_STREAM_DRAW);
	pboMemory=glext->glMapBuffer(PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY);
	if(p->components!=3)
	  memcpy(pboMemory, texData, alloc_size);
	else {
	  int i, max=rw*rh;
	  char *p1=(char *)pboMemory, *p2=(char *)texData;
	  for(i=max; i>0; i--) {
	    *p1++=p2[2];
	    *p1++=p2[1];
	    *p1++=p2[0];
	    *p1++=0;
	    p2+=3;
	  }
	}
	glext->glUnmapBuffer(PIXEL_UNPACK_BUFFER_EXT);
	if(format!=GL_LUMINANCE)
	  format=GL_BGRA;
	printf("out_space=%s, rw=%d, rh=%d, format=%s\n", get_color_space_string(out_space), rw, rh, get_color_space_string(format));
	glTexImage2D(texture_target, 0, out_space, rw, rh, 0, format, type, 0);
	glext->glDeleteBuffers(1, &buf);
      }
    }
  }
  glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
  glPixelStorei(GL_UNPACK_IMAGE_HEIGHT, 0);

  glFinish();
  GLenum e;
  if((e=glGetError())!=GL_NO_ERROR) {
    printf("GL Error: %s\n", gluErrorString(e));
    //loaded=false;
    //return;
  }

  set_ratio(((double)p->w)/p->h);

  if(glext->nv_tex_rect_supported) {
    if(colorspace==PIXMAP_COLORSPACE_YUV_COMPLEX) {
      for(int i=0; i<3; i++) {
	maxx[i]=p->allocw_per_channel[i];
	maxy[i]=p->alloch_per_channel[i]*p->iterations;
      }
    } else {
      maxx[0]=p->w, maxy[0]=p->h;
    }
  } else {
    maxx[0]=((double)p->w)/rw;
    maxy[0]=((double)p->h)/rh;
  }

  loaded=true;
}

void GLSimplePixmap::raw_display() // display texture provided that color space conversion is done
{
  double x1=-1, x2=-1, x3=1, x4=1, y1=1, y2=-1, y3=-1, y4=1, tmp, tmp2;
  if(flip)
    tmp=y1, y1=y2, y2=tmp, tmp=y3, y3=y4, y4=tmp;
  for(int i=0; i<rotate; i++) {
    tmp=x2, tmp2=y2;
    x2=x1, y2=y1;
    x1=tmp, y1=tmp2;

    tmp=x3, tmp2=y3;
    x3=x1, y3=y1;
    x1=tmp, y1=tmp2;

    tmp=x4, tmp2=y4;
    x4=x1, y4=y1;
    x1=tmp, y1=tmp2;
  }

  if(colorspace==PIXMAP_COLORSPACE_YUV_COMPLEX) {
    glext->glActiveTexture(GL_TEXTURE2);
    glEnable(texture_target);
    glBindTexture(texture_target, texture[2]);
    glext->glActiveTexture(GL_TEXTURE1);
    glEnable(texture_target);
    glBindTexture(texture_target, texture[1]);
    glext->glActiveTexture(GL_TEXTURE0);
    glBindTexture(texture_target, texture[0]);
    glBegin(GL_QUADS);
    for(int i=0; i<3; i++)
      glext->glMultiTexCoord2f(i, 0, 0);
    glVertex2f(x1, y1);
    for(int i=0; i<3; i++)
      glext->glMultiTexCoord2f(GL_TEXTURE0+i, 0, maxy[i]);
    glVertex2f(x2, y2);
    for(int i=0; i<3; i++)
      glext->glMultiTexCoord2f(GL_TEXTURE0+i, maxx[i], maxy[i]);
    glVertex2f(x3, y3);
    for(int i=0; i<3; i++)
      glext->glMultiTexCoord2f(GL_TEXTURE0+i, maxx[i], 0);
    glVertex2f(x4, y4);
    glEnd();
    glext->glActiveTexture(GL_TEXTURE2);
    glDisable(texture_target);
    glext->glActiveTexture(GL_TEXTURE1);
    glDisable(texture_target);
    glext->glActiveTexture(GL_TEXTURE0);
  } else {
    glBindTexture(texture_target, texture[0]);
    glBegin(GL_QUADS);
    glTexCoord2f(0, 0);
    glVertex2f(x1, y1);
    glTexCoord2f(0, maxy[0]);
    glVertex2f(x2, y2);
    glTexCoord2f(maxx[0], maxy[0]);
    glVertex2f(x3, y3);
    glTexCoord2f(maxx[0], 0);
    glVertex2f(x4, y4);
    glEnd();
  }
}

static void _setup_register_combiners_yuv()
{
  glEnable(GL_REGISTER_COMBINERS_NV);

  // based on code of Reimar Dffinger from mplayer g2 vo_gl
  GLfloat ucoef[] = { 0, -0.344, 1.770, 0 };
  GLfloat vcoef[] = { 1.403, -0.714, 0, 0 };

  // It seems that coefficient must be in [0,1] range, so here comes
  // the trick:
  // first put them in [-0.5, 0.5] range, then add 0.5;
  // can be undone with the HALF_BIAS and SCALE_BY_FOUR arguments
  // for CombinerInput and CombinerOutput
  for (int i = 0; i < 4; i++) {
    ucoef[i] = ucoef[i] * 0.25 + 0.5;
    vcoef[i] = vcoef[i] * 0.25 + 0.5;
  }

  glext->glCombinerParameterfv(GL_CONSTANT_COLOR0_NV, ucoef);
  glext->glCombinerParameterfv(GL_CONSTANT_COLOR1_NV, vcoef);
  if(glGetError()!=GL_NO_ERROR) {
    printf("opengl error\n");
    exit(1);
  }

  //separate YUV channels: Y->SPARE0, U->SPARE1, V->SPARE0 alpha
  // UV first, like this green component can't overflow
  // separating Y and U with primary color (red) and secondary color(blue)

  glext->glCombinerInput(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_A_NV,
      GL_TEXTURE0, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glext->glCombinerInput(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_B_NV,
      GL_SECONDARY_COLOR_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glext->glCombinerInput(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_C_NV,
      GL_TEXTURE0, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glext->glCombinerInput(GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_D_NV,
      GL_FOG, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glext->glCombinerOutput (GL_COMBINER0_NV, GL_RGB, GL_SPARE0_NV,
      GL_SPARE1_NV, GL_DISCARD_NV,
      GL_NONE, GL_NONE, GL_TRUE,
      GL_TRUE, GL_FALSE);

  glext->glCombinerInput(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_A_NV,
      GL_TEXTURE0, GL_UNSIGNED_IDENTITY_NV, GL_BLUE);
  glext->glCombinerInput(GL_COMBINER0_NV, GL_ALPHA, GL_VARIABLE_B_NV,
      GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_ALPHA);
  glext->glCombinerOutput (GL_COMBINER0_NV, GL_ALPHA, GL_SPARE0_NV,
      GL_DISCARD_NV, GL_DISCARD_NV,
      GL_NONE, GL_NONE, GL_FALSE,
      GL_FALSE, GL_FALSE);

  glext->glCombinerInput(GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_A_NV,
      GL_CONSTANT_COLOR0_NV, GL_HALF_BIAS_NORMAL_NV, GL_RGB);
  glext->glCombinerInput (GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_A_NV,
      GL_SPARE1_NV, GL_HALF_BIAS_NORMAL_NV, GL_RGB);
  glext->glCombinerInput (GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_B_NV,
      GL_CONSTANT_COLOR0_NV, GL_HALF_BIAS_NORMAL_NV, GL_RGB);
  glext->glCombinerInput (GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_C_NV,
      GL_SPARE0_NV, GL_HALF_BIAS_NORMAL_NV, GL_ALPHA);
  glext->glCombinerInput (GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_D_NV,
      GL_CONSTANT_COLOR1_NV, GL_HALF_BIAS_NORMAL_NV, GL_RGB);
  glext->glCombinerOutput (GL_COMBINER1_NV, GL_RGB, GL_DISCARD_NV,
      GL_DISCARD_NV, GL_SPARE1_NV,
      GL_SCALE_BY_FOUR_NV, GL_NONE, GL_FALSE,
      GL_FALSE, GL_FALSE);

  glext->glCombinerInput (GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_A_NV,
      GL_SPARE0_NV, GL_SIGNED_IDENTITY_NV, GL_RGB);
  glext->glCombinerInput (GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_B_NV,
      GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
  glext->glCombinerInput (GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_C_NV,
      GL_SPARE1_NV, GL_SIGNED_IDENTITY_NV, GL_RGB);
  glext->glCombinerInput (GL_COMBINER2_NV, GL_RGB, GL_VARIABLE_D_NV,
      GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);
  glext->glCombinerOutput (GL_COMBINER2_NV, GL_RGB, GL_DISCARD_NV,
      GL_DISCARD_NV, GL_SPARE0_NV, GL_NONE,
      GL_NONE, GL_FALSE, GL_FALSE, GL_FALSE);


  glext->glCombinerParameteri(GL_NUM_GENERAL_COMBINERS_NV, 3);

  glEnable(GL_FOG);
  glext->glSecondaryColor3f(1, 0, 0);
  glext->glActiveTexture(GL_TEXTURE0);
  float green[]={0, 1, 0, 0};
  glFogfv(GL_FOG_COLOR, green);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
  glext->glFinalCombinerInput(GL_VARIABLE_A_NV, GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glext->glFinalCombinerInput(GL_VARIABLE_B_NV, GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glext->glFinalCombinerInput(GL_VARIABLE_C_NV, GL_ZERO, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
  glext->glFinalCombinerInput(GL_VARIABLE_D_NV, GL_SPARE0_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
}

static void _setup_register_combiners_yuv_3tex()
{
  glEnable(GL_REGISTER_COMBINERS_NV);

  // based on code of Reimar Dffinger from mplayer g2 vo_gl
  GLfloat ucoef[] = { 0, -0.344, 1.770, 0 };
  GLfloat vcoef[] = { 1.403, -0.714, 0, 0 };

  // It seems that coefficient must be in [0,1] range, so here comes
  // the trick:
  // first put them in [-0.5, 0.5] range, then add 0.5;
  // can be undone with the HALF_BIAS and SCALE_BY_FOUR arguments
  // for CombinerInput and CombinerOutput
  for (int i = 0; i < 4; i++) {
    ucoef[i] = ucoef[i] * 0.25 + 0.5;
    vcoef[i] = vcoef[i] * 0.25 + 0.5;
  }

  glext->glCombinerParameterfv(GL_CONSTANT_COLOR0_NV, ucoef);
  glext->glCombinerParameterfv(GL_CONSTANT_COLOR1_NV, vcoef);
  if(glGetError()!=GL_NO_ERROR) {
    printf("opengl error\n");
    exit(1);
  }

  // UV first, like this green component can't overflow                                                                                  
  glext->glCombinerInput (GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_A_NV,                                                                  
      GL_TEXTURE1, GL_HALF_BIAS_NORMAL_NV, GL_RGB);                                                               
  glext->glCombinerInput (GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_B_NV,                                                                  
      GL_CONSTANT_COLOR0_NV,                                                                                      
      GL_HALF_BIAS_NORMAL_NV, GL_RGB);                                                                            
  glext->glCombinerInput (GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_C_NV,                                                                  
      GL_TEXTURE2, GL_HALF_BIAS_NORMAL_NV, GL_RGB);                                                               
  glext->glCombinerInput (GL_COMBINER0_NV, GL_RGB, GL_VARIABLE_D_NV,                                                                  
      GL_CONSTANT_COLOR1_NV,                                                                                      
      GL_HALF_BIAS_NORMAL_NV, GL_RGB);                                                                            
  glext->glCombinerOutput (GL_COMBINER0_NV, GL_RGB, GL_DISCARD_NV,                                                                    
      GL_DISCARD_NV, GL_SPARE0_NV,                                                                               
      GL_SCALE_BY_FOUR_NV, GL_NONE, GL_FALSE,                                                                    
      GL_FALSE, GL_FALSE);                                                                                       

  glext->glCombinerInput (GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_A_NV,                                                                  
      GL_SPARE0_NV, GL_SIGNED_IDENTITY_NV, GL_RGB);                                                               
  glext->glCombinerInput (GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_B_NV,                                                                  
      GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);                                                                    
  glext->glCombinerInput (GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_C_NV,                                                                  
      GL_TEXTURE0, GL_SIGNED_IDENTITY_NV, GL_RGB);                                                                
  glext->glCombinerInput (GL_COMBINER1_NV, GL_RGB, GL_VARIABLE_D_NV,                                                                  
      GL_ZERO, GL_UNSIGNED_INVERT_NV, GL_RGB);                                                                    
  glext->glCombinerOutput (GL_COMBINER1_NV, GL_RGB, GL_DISCARD_NV,                                                                    
      GL_DISCARD_NV, GL_SPARE0_NV, GL_NONE,                                                                      
      GL_NONE, GL_FALSE, GL_FALSE, GL_FALSE);                                                                    

  // leave final combiner stage in default mode                                                                                          
  glext->glCombinerParameteri (GL_NUM_GENERAL_COMBINERS_NV, 2);                                  

  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}
static void _disable_register_combiners()
{
  glDisable(GL_FOG);
  glDisable(GL_REGISTER_COMBINERS_NV);
}

void GLSimplePixmap::display()
{
  if(colorspace==PIXMAP_COLORSPACE_YUV) {
    glext->glActiveTexture(GL_TEXTURE0);
    _setup_register_combiners_yuv();
  } else if(colorspace==PIXMAP_COLORSPACE_YUV_COMPLEX) {
    glext->glActiveTexture(GL_TEXTURE0);
    _setup_register_combiners_yuv_3tex();
  } else {
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
  }
  raw_display();
  if(colorspace==PIXMAP_COLORSPACE_YUV||colorspace==PIXMAP_COLORSPACE_YUV_COMPLEX)
    _disable_register_combiners();
}
GLSimplePixmap::~GLSimplePixmap()
{

  if(colorspace==PIXMAP_COLORSPACE_YUV_COMPLEX) {
    printf("deleting textures %d, %d, %d\n", texture[0], texture[1], texture[2]);
    glDeleteTextures(3, texture);
  } else {
    printf("deleting texture %d\n", texture[0]);
    glDeleteTextures(1, texture);
  }
}
