#include "conf.hh"
#include <stdio.h>
#include <string.h>
#ifdef USESDL
#include <SDL/SDL.h>
#else
#include <GL/glut.h>
#endif
#include "glext.hh"
#include <stdlib.h>
#include <time.h>
#include "view.hh"


int win;
clock_t start;
View *v;
void timer_start()
{
  start=clock();
}
double timer_get()
{
  return(((double)clock()-start)/CLOCKS_PER_SEC);
}
void disp()
{
  v->display();
}
void keyb(unsigned char key, int x, int y)
{
  v->keyb(key, x, y);
}
#ifdef USESDL
void sdl_quit(int code)
{
  SDL_Quit();
  exit(0);
}

static void sdl_process_events( void )
{
  SDL_Event event;
  while( SDL_PollEvent( &event ) ) {
    switch( event.type ) {
      case SDL_KEYDOWN:
	v->keyb(event.key.keysym.sym, 0, 0);
	break;
      case SDL_QUIT:
	sdl_quit( 0 );
	break;
    }
  }
}
int width, height;
extern "C" {
  SDL_GrabMode SDL_WM_GrabInputOff();
}
void sdl_init()
{
  SDL_Surface *screen;
  const SDL_VideoInfo* info = NULL;
  if( SDL_Init( SDL_INIT_VIDEO ) < 0 ) {
    fprintf( stderr, "Video initialization failed: %s\n",
	SDL_GetError( ) );
    sdl_quit( 1 );
  }
  info = SDL_GetVideoInfo( );

  if( !info ) {
    fprintf( stderr, "Video query failed: %s\n", SDL_GetError( ) );
    sdl_quit( 1 );
  }
  int width = 1280;
  int height = 1024;
  int bpp = info->vfmt->BitsPerPixel;
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
  int flags = SDL_OPENGL|SDL_FULLSCREEN;
  if( (screen=SDL_SetVideoMode( width, height, bpp, flags )) == 0 ) {
    fprintf( stderr, "Video mode set failed: %s\n", SDL_GetError( ) );
    sdl_quit( 1 );
  }
  //printf("%d\n", SDL_WM_GrabInput((SDL_GrabMode)-3));
  SDL_WM_GrabInputOff();
  //SDL_Quit();
  //exit(0);
  glViewport( 0, 0, width, height);

}
void sdl_main()
{
  while(1) {
    sdl_process_events();
    disp();
  }
}
#else
void glut_init(int *argc, char **argv)
{
  glutInit(argc, argv);
  glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE);
  win=glutCreateWindow("glsee");
  glutFullScreen();
  glutDisplayFunc(disp);
  glutKeyboardFunc(keyb);
}
#endif
void quit()
{
#ifdef USESDL
  SDL_Quit();
#else
  glutDestroyWindow(win);
#endif
 delete v;
 exit(0);
}
void swap_buffers()
{
#ifdef USESDL
  glLoadIdentity();
  SDL_GL_SwapBuffers();
  glFinish();
#else
  glLoadIdentity();
  glutSwapBuffers();
#endif
}

void gl_init()
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_DEPTH_TEST);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-1,1,-1,1,-10,10);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

static void setup_opengl( int width, int height )
{
    glClearColor( 0, 0, 0, 0 );

    glViewport( 0, 0, width, height );

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );
    glOrtho(-2, 2, -2, 2, -10, 10);
}

main(int argc, char **argv)
{
  int i;
  argv++;
  argc--;
  if(argc<1) {
    exit(0);
  }
#ifdef USESDL
  sdl_init();
#else
  glut_init(&argc, argv);
#endif
  glext=new GLExt();
  /*if(argc>=3&&!strcmp(argv[0], "-tjl")) {
    test_jpegload(argv[1], atoi(argv[2]));
    return 0;
  }
  if(argc>=3&&!strcmp(argv[0], "-tgl")) {
    test_glpload(argv[1], atoi(argv[2]));
    return 0;
  }
  */
  URIContainerTree *doc=new URIContainerTree();
  for(i=0; i<argc; i++)
    doc->add(argv[i]);
  vector<int> p;
  /*for(i=0; i<argc; i++) {
    ViewableDocument *z;
    if((z=doc->prev(&p, 0))==NULL) {
      printf("ERROR: not good documents\n");
      exit(1);
    }
    printf("%d. document: %s\n", i, z->uri);
  }
  printf("pathlength: %d\n", p.size());
  if(doc->prev(&p, 0)!=NULL) {
    printf("ERROR: too much documents\n");
    exit(1);
  }
  */
  v=new View(doc);
  gl_init();
  v->init_coords();
  v->display();
  v->prepare();

#ifdef USESDL
  sdl_main();
#else
  glutMainLoop();
#endif
  return 0;
}




