#include <fcntl.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <GL/gl.h>

#include "conf.hh"
#include "document.hh"
#include "pixmap.hh"
#include "glext.hh"

class PSDocument : public ViewableDocument {
    PSDocumentTree *parent;
    int page, start, len;
  public:
    PSDocument(PSDocumentTree *parent, int page, int start, int len);
    Pixmap *get_pixmap();
    void init();
};

class PSPixmap : public Pixmap {
  public:
    PSPixmap(PSDocumentTree *parent, int page, int start, int len);
};

PSDocument::PSDocument(PSDocumentTree *parent, int page, int start, int len) :
  parent(parent), page(page), start(start), len(len)
{
  printf("page %d starts@%d, length: %d\n", page, start, len);
}
void PSDocument::init()
{
  if(!flags&DOCUMENT_TREE_INITED) {
    flags|=DOCUMENT_TREE_INITED;
  }
}
Pixmap *PSDocument::get_pixmap()
{
  if(p)
    return p;
  p=new PSPixmap(parent, page, start, len);
  printf("page loaded, p=%x, p->data=%x\n", p, p->data);
  return p;
}

PSDocumentTree::PSDocumentTree(char *uri0)
{
  flags&=~PSDOCTREE_TMPURI;
#ifndef USE_ACROREAD
  if(has_mimetype(uri0, "pdf"))
#else
  if(0)
#endif
  {
    char *s;
    int z;
    uri=strdup("/tmp/pdfXXXXXX");
    z=mkstemp(uri);
    printf("pdf uri: %s\n", uri);
    close(z);
    flags|=PSDOCTREE_TMPURI;
    asprintf(&s, "pdf2dsc '%s' '%s'\n", uri0, uri);
    system(s);
    free(s);
  } else {
    uri=strdup(uri0);
  }
    uri=strdup(uri0);
  children=new vector<DocumentTree *>;
  child=-1;
}
PSDocumentTree::~PSDocumentTree()
{
  if(child>=0) {
    printf("killing child\n");
    write(wfd, "quit\n", 5);
    close(rfd);
    close(wfd);
    printf("killing child %d\n");
    kill(child, SIGTERM);
  }
  if(flags&PSDOCTREE_TMPURI)
    unlink(uri);
}
int callacroread(char *uri)
{
  int r;
  int p[2];
  pipe(p);
  r=p[0];
  printf("calling acroread\n");
  int pid=fork();
  if(pid==0) {
    char *s;
    asprintf(&s, "cat '%s'|acroread -toPostScript -fast", uri);
    printf("really calling with %s\n", s);
    //close(0);
    close(1);
    //close(2);
    dup2(p[1], 1);
    system(s);
    free(s);
    fprintf(stderr, "acroread exited\n");
    exit(0);
  }
  close(p[1]);
  printf("acroread called\n");
  return(r);
}
char *loadtostring(char *uri)
{
  char *r=(char *)malloc(4096);
  int h;
  int l=0, alloced=0, br=1;
  printf("loading %s\n", uri);
#ifdef USE_ACROREAD
  if(has_mimetype(uri, "pdf"))
#else
  if(0)
#endif
    h=callacroread(uri);
  else
    h=open(uri, O_RDONLY);
  printf("reading...\n");
  while(br>0) {
    if(alloced-l<4096) {
      alloced=l+4096;
      r=(char *)realloc(r, alloced);
    }
    br=read(h, &r[l], 4096);
    if(br>0)
      l+=br;
  }
  r[l]=0;
  close(h);
  printf("%s loaded\n", uri);
  return r;
}
int callgs(int *rfd, int *wfd)
{
  int p1[2], p2[2];
  pipe(p1);
  pipe(p2);
  *rfd=p1[0];
  *wfd=p2[1];
  printf("calling gs\n");
  int r=fork();
  if(r==0) {
    printf("really calling\n");
    close(0);
    close(1);
    close(2);
    dup2(p2[0], 0);
    dup2(p1[1], 1);
    char *argv[]={"gs", "-q", "-", NULL};
    //execve("gs", argv, environ);
    system("gs -r200 -q -sOutputFile=- -sDEVICE=pgmraw -");
    exit(0);
  }
  printf("gs called\n");
  return(r);
}
void PSDocumentTree::init()
{
  if(!(flags&DOCUMENT_TREE_INITED)) {
    flags|=DOCUMENT_TREE_INITED;

    printf("loading ps images\n");
    psfile=loadtostring(uri);
    char *p=psfile;
    int last=0;
    int i=0, startlen;
    printf("ps loaded\n");
    //printf("loaded string: %s\n", p);
    while((p=strstr(p, "%%"))!=NULL) {
      if(!strncmp(p, "%%Page:", 7)) {
	printf("page found\n");
	if(last==0)
	  startlen=p-psfile;
	else
	  children->push_back(new PSDocument(this, ++i, last, p-psfile-last));
	last=p-psfile;
      }
      if(!strncmp(p, "%%Trailer", 9)) {
	printf("this is the last page\n");
	children->push_back(new PSDocument(this, ++i, last, p-psfile-last));
	break;
      }
      p++;
    }
    child=callgs(&rfd, &wfd);
    write(wfd, psfile, startlen);
  }
}
char *readln(int fd)
{
  char *s=NULL;
  int z;
  int l=0;
  do {
    s=(char *)realloc(s, l+1);
    z=read(fd, &s[l], 1);
    l++;
  } while(z==1&&s[l-1]!='\n');
  s[l-1]=0;
  return s;
}
int mustread(int fd, unsigned char *data, int l)
{
  int r=1, done=0;
  while(r>0&&done<l) {
    r=read(fd, &data[done], l-done);
    done+=r;
  }
  return done;
}
PSPixmap::PSPixmap(PSDocumentTree *doc, int page, int start, int len)
{
  int format;
  printf("showing page %d\n");
  write(doc->wfd, &doc->psfile[start], len);
  char *s;
  s=readln(doc->rfd);
  printf("%s\n", s);
  format=5;
  if(!strcmp(s, "P6"))
    format=6;
  else if(!strcmp(s, "P5"))
    format=5;
  else {
    image_valid=false;
    data=NULL;
    return;
  }
  s=readln(doc->rfd);
  printf("%s\n", s);
  s=readln(doc->rfd);
  printf("%s\n", s);
  sscanf(s, "%d %d", &w, &h);
  flip=0;
  read(doc->rfd,data,1);
  if(format==6) 
    colorspace=GL_RGB;
  else if(format==5)
    colorspace=GL_LUMINANCE;
  data=(unsigned char *)malloc(alloc_size);
  int z=mustread(doc->rfd, data, w*h*components);
}
