#include "conf.hh"
#ifdef HAVE_TIFF
#include "document.hh"

#include "tiffio.h"

#include <locale.h>
#include <stdio.h>
#include <stdlib.h>

#include <GL/gl.h>


#include "pixmap.hh"
#include "glext.hh"

class TIFFPixmap : public Pixmap {
  public:
    TIFFPixmap(TIFFDocumentTree *parent, int page);
};

class TIFFDocument : public ViewableDocument {
    TIFFDocumentTree *parent;
    int page;
  public:
    TIFFDocument(TIFFDocumentTree *parent, int page);
    Pixmap *get_pixmap();
    void init();
};

TIFFDocument::TIFFDocument(TIFFDocumentTree *parent, int page) : parent(parent), page(page)
{
  printf("d");
}
void TIFFDocument::init()
{
  if(!flags&DOCUMENT_TREE_INITED) {
    flags|=DOCUMENT_TREE_INITED;
  }
}
Pixmap *TIFFDocument::get_pixmap()
{
  if(p)
    return p;
  p=new TIFFPixmap(parent, page);
  printf("page loaded, p=%x, p->data=%x\n", p, p->data);
  return p;
}

TIFFDocumentTree::TIFFDocumentTree(char *uri0)
{
  uri=strdup(uri0);
  children=new vector<DocumentTree *>;
}
void TIFFDocumentTree::init()
{
  if(!flags&DOCUMENT_TREE_INITED) {
    flags|=DOCUMENT_TREE_INITED;

    printf("loading TIFF images\n");

    doc=TIFFOpen(uri, "r");
    if (! doc) {
      printf("couldn't open TIFF file %s\n", uri);
      return;
    }

    while(TIFFReadDirectory(doc)==1);

    int p=TIFFCurrentDirectory(doc);
    printf("%d pages found\n", p);
    for(int i=0; i<p; i++)
      children->push_back(new TIFFDocument(this, i));
  }
}
TIFFPixmap::TIFFPixmap(TIFFDocumentTree *doc, int page_num)
{
    printf("loading page %d\n", page_num);
    TIFF *tif=doc->doc;
    TIFFSetDirectory(tif, page_num);
    flip=false;
    tsample_t spp, bps;

    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
    TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &spp);
    TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bps);

    printf("tiff w=%d, h=%d, spp=%d, bps=%d\n", w, h, spp, bps);
    data=NULL;
    if(spp==1) {
      colorspace=GL_LUMINANCE;
      compute_alloc_size();
      data=(unsigned char *)malloc(alloc_size);
      data[0]=0;
      data[1]=0xff;
      data[2]=0xff;
      data[3]=0xff;
      data[4]=0xff;
      data[5]=0xff;
      
      //colorspace=GL_BITMAP;
      image_valid=true;
    } else {
      printf("Sorry, currently only bitmaps are supported in TIFFs.");
      image_valid=false;
    }
}
#endif
