#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

#include "view.hh"
void View::fit_width()
{
  if(!curdoc)
    return;
  GLPixmap *glp=curdoc->get_glp();
  if(!glp)
    return;
  double ratio=glp->get_ratio();
  if(rotate&1)
    ratio=1/ratio;
  double sratio=ratio/1.25;

  sx=0, sy=0, sw=1; sh=sw/sratio;
  if(sh<1)
    sy=(1-sh)/2;

}
void View::fit_height()
{
  if(!curdoc)
    return;
  GLPixmap *glp=curdoc->get_glp();
  if(!glp)
    return;
  double sratio=glp->get_ratio()/1.25;
  sx=0, sy=0, sh=1; sw=sh*sratio;
  if(sw<1)
    sx=(1-sw)/2;
}
void View::fit_all()
{
  if(!curdoc)
    return;
  GLPixmap *glp=curdoc->get_glp();
  if(!glp)
    return;
  double sratio=glp->get_ratio()/1.25;
  sx=0, sy=0, sh=1; sw=sh*sratio;
  if(sw>1)
    sw=1, sh=sw/sratio;
  sx=(1-sw)/2;
  sy=(1-sh)/2;
}
bool View::init_coords()
{
  if(!curdoc)
    return false;
  //fit_width();
  fit_all();
  if(curdoc->glp)
    curdoc->glp->rotate=rotate;
  return true;
}
#define MAXSCOPE 300
void View::prepare()
{
  vector<int> nextpath=*curpath;
  ViewableDocument *v;
  v=docs->next(&nextpath, 0);
  if(v) {
    v->get_glp();
    printf("[next=%x]\n", v);
    for(int i=0;v&&(i<10);v=docs->next(&nextpath, 0), i++);
    printf("[nnext=%x]\n", v);
    if(v)
      v->del_glp();
  }
  vector<int> prevpath=*curpath;
  v=docs->prev(&prevpath, 0);
  if(v) {
    v->get_glp();
    printf("[prev=%x]\n", v);
    for(int i=0;v&&(i<10);v=docs->prev(&prevpath, 0), i++);
    printf("[pprev=%x]\n", v);
    if(v)
      v->del_glp();
  }
}
void View::next_doc()
{
//  if(curdoci<docs.size()-1) {
//   if(curdoci>MAXSCOPE)
//      docs[curdoci-MAXSCOPE]->del_glp();
//    curdoci++;
  printf("calling next...\n");
  vector<int> p(*curpath);
  if(docs->next(&p,0)) {
    curdoc=docs->next(curpath, 0);
    printf("done\n");
    if(curdoc) {
      init_coords();
      display();
      glFinish();
      prepare();
    }
  }
}
void View::prev_doc()
{
  //if(curdoci>0) {
   // if(docs.size()>curdoci+MAXSCOPE)
   //   docs[curdoci+MAXSCOPE]->del_glp();
   // curdoci--;
  printf("calling prev...\n");
  vector<int> p(*curpath);
  if(docs->prev(&p,0)) {
    curdoc=docs->prev(curpath, 0);
    printf("done\n");
    if(curdoc) {
      init_coords();
      display();
      glFinish();
      prepare();
    }
  }
}
void View::display()
{
  if(!curdoc)
    return;
  GLPixmap *glp=curdoc->get_glp();
  if(!glp)
    return;
  double ratio=glp->get_ratio();
  glClear(GL_COLOR_BUFFER_BIT);
  printf("x: %f, y: %f, w: %f, h: %f\n", sx, sy, sw, sh);

  glPushMatrix();
  glTranslatef((sx+sw)-1, (-sh+1)-2*sy, 0);
  glScalef(sw, sh, 1);
  glp->display();
  glPopMatrix();

  swap_buffers();
}
void View::keyb(unsigned char key, int x, int y)
{
  switch(key) {
    case 'q':
      quit();
      break;
    case 'p':
      prev_doc();
      break;
    case 'n':
      next_doc();
      break;
    case 'j':
      printf("sh=%f, sy=%f\n", sh, sy);
      if(sh>=1&&sy+sh>1) {
	sy-=0.2;
	if(sy+sh<1)
	  sy=1-sh;
	display();
      }
      break;
    case 'l':
      printf("l: sx=%f, sw=%f\n", sx, sw);
      if(sx+sw*4/3>1) {
	sx-=0.2;
	if(sx+sw*4/3<1)
	  sx=1-sw*4/3;
	display();
      }
      break;
    case 'h':
      if(sx<0) {
	sx+=0.2;
	if(sx>0)
	  sx=0;
	display();
      }
      break;
    case 'u':
      if(sh>=1&&sy<0) {
	sy+=0.7;
	if(sy>0)
	  sy=0;
	display();
      }
      break;
    case 'd':
      if(sh>=1&&sy+sh>1) {
	sy-=0.7;
	if(sy+sh<1)
	  sy=1-sh;
	display();
      }
      break;
    case 'k':
      if(sh>=1&&sy<0) {
	sy+=0.2;
	if(sy>0)
	  sy=0;
	display();
      }
      break;
    case '.':
      sx=1-(1-sx)*1.5;
      sy=0.5-(0.5-sy)*1.5;
      sh*=1.5;
      sw*=1.5;
      display();
      break;
    case ',':
      sx=1-(1-sx)/1.5;
      sy=0.5-(0.5-sy)/1.5;
      sh/=1.5;
      sw/=1.5;
      display();
      break;
    case 'x':
      if(!mkdir(".deleted", 0777)||errno==EEXIST) {
	char *s;
	asprintf(&s, ".deleted/%s", curdoc->uri);
	printf("moving %s to %s\n", curdoc->uri, s);
	rename(curdoc->uri, s);
	free(s);
      }
      break;
    case ' ':
      if(sh>=1&&sy+sh>1) {
	sy-=sh;
	if(sy+sh<1)
	  sy=1-sh;
	display();
      } else
	next_doc();
      break;
    case 'e':
      rotate++;
      if(rotate>3)
	rotate=0;

      curdoc->get_glp()->rotate++;
      if(curdoc->get_glp()->rotate>3)
	curdoc->get_glp()->rotate=0;
      init_coords();
      display();
      break;
    case 'w':
      rotate--;
      if(rotate<0)
	rotate=3;

      curdoc->get_glp()->rotate--;
      if(curdoc->get_glp()->rotate<0)
	curdoc->get_glp()->rotate=3;
      init_coords();
      display();
      break;
      
  }
}
View::View(DocumentTree *docs0)
{
  docs=docs0;
  curpath=new vector<int>;
  printf("docs: %x\n", docs0);
  curdoc=docs->next(curpath, 0);
  rotate=0;
}
View::~View()
{
  printf("deleting view\n");
  //for(int i=0; i<docs.size(); i++)
  //  delete docs[i];
  delete curpath;
  delete docs;
}
